@@ -88,6 +88,9 @@
 #define countof(x)	(sizeof(x) / sizeof(x[0]))
 #endif
 
+#ifndef howmany
+#define howmany(x, y)	(((x) + (y) - 1) / (y))
+#endif
 #ifndef roundup
 #define roundup(x, y)	((((x) + ((y) - 1)) / (y)) * (y))
 #endif
