@@ -32,6 +32,29 @@
 #include "image_priv.h"
 #include <string.h>
 
+//#define IMAGE_PROFILE
+
+#if defined(IMAGE_PROFILE)
+#include <time.h>
+#define PROF(x)	clock_gettime(CLOCK_MONOTONIC, &x);
+#define PROF_RESULT(msg, x)	do {	\
+	uint32 x##_us = timespec_to_usec(&x##_end) - timespec_to_usec(&x##_start);\
+	printf("%-12s %u.%06u sec\n", msg, x##_us / 1000000, x##_us % 1000000);	\
+} while (0)
+#else
+#define PROF(x)	/**/
+#define PROF_RESULT(msg, x)	/**/
+#endif
+
+// r, g, b を NTSC 輝度に変換する。高速化のため雑に近似。
+// 元の式は (0.299 * R) + (0.587 * G) + (0.114 * B) で、
+// これを整数演算にした (77 * R + 150 * G + 29 * B) / 256 とかが定番だが、
+// gcc(m68k) だと乗算命令使わず筆算式に展開するっぽくそれなら立ってるビットが
+// 少ないほうがいいので 4 ビットで済ませる。
+// 元々この係数の厳密性はそんなにないので、この程度で十分。
+// 普通に乗算命令を吐くアーキテクチャ/コンパイラなら、どっちでも一緒。
+#define RGBToY(r, g, b) ((5 * (r) + 9 * (g) + 2 * (b)) / 16)
+
 struct image_reductor_handle_;
 typedef uint (*finder_t)(struct image_reductor_handle_ *, ColorRGB);
 
@@ -49,43 +72,49 @@ typedef struct {
 
 typedef struct image_reductor_handle_
 {
+	// 画像 (所有はしていない)
+	struct image *dstimg;
+	struct image *srcimg;
+
 	bool is_gray;
 
-	uint gain;
+	// ゲイン。256 を 1.0 とする。負数なら適用しない (1.0 のまま)。
+	int gain;
 
 	// 色からパレット番号を検索する関数。
 	finder_t finder;
 
-	// パレット (palette_buf か固定パレットを指す)
-	const ColorRGB *palette;
-	uint palette_count;
+	// RGB555 から適応パレットのカラーコードを引くハッシュ。
+	uint16 *colorhash;
 
-	// 動的に作成する場合はここがメモリを所有している。
-	ColorRGB *palette_buf;
+	// 適応パレット時に使う検索範囲の上下限。
+	int y_lo[8];
+	int y_hi[8];
 } image_reductor_handle;
 
 static uint finder_gray(image_reductor_handle *, ColorRGB);
 static uint finder_fixed8(image_reductor_handle *, ColorRGB);
 static uint finder_vga16(image_reductor_handle *, ColorRGB);
-static uint finder_fixed256(image_reductor_handle *, ColorRGB);
 #if defined(SIXELV)
+static uint finder_fixed256(image_reductor_handle *, ColorRGB);
 static uint finder_xterm256(image_reductor_handle *, ColorRGB);
 static inline uint8 finder_xterm256_channel(uint8);
 #endif
+static uint finder_adaptive(image_reductor_handle *, ColorRGB);
 static void colorcvt_gray(ColorRGBint32 *);
 static ColorRGB *image_alloc_gray_palette(uint);
-static ColorRGB *image_alloc_fixed256_palette(void);
 #if defined(SIXELV)
+static ColorRGB *image_alloc_fixed256_palette(void);
 static ColorRGB *image_alloc_xterm256_palette(void);
 #endif
+static bool image_calc_adaptive_palette(image_reductor_handle *);
 
 #if defined(SIXELV)
-static void image_reduct_simple(image_reductor_handle *,
-	struct image *, const struct image *, const struct diag *);
+static bool image_reduct_simple(image_reductor_handle *,
+	const struct image_opt *, const struct diag *);
 #endif
 static bool image_reduct_highquality(image_reductor_handle *,
-	struct image *, const struct image *, const image_opt *,
-	const struct diag *);
+	const struct image_opt *, const struct diag *);
 #if defined(SIXELV)
 static void set_err(ColorRGBint16 *, int, const ColorRGBint32 *, int);
 #endif
@@ -96,21 +125,83 @@ static inline int16 saturate_adderr(int16, int);
 
 static const ColorRGB palette_fixed8[];
 static const ColorRGB palette_vga16[];
+#if defined(IMAGE_PROFILE)
+static int y_count[8];
+#endif
 
 // opt を初期化する。
 void
-image_opt_init(image_opt *opt)
+image_opt_init(struct image_opt *opt)
 {
 	opt->method  = REDUCT_HIGH_QUALITY;
 	opt->diffuse = DIFFUSE_SFL;
-	opt->color   = COLOR_FMT_256_RGB332;
+	opt->color   = MAKE_COLOR_MODE_ADAPTIVE(256);
 	opt->cdm     = 0;
-	opt->gain    = 256;
+	opt->gain    = -1;
 	opt->output_ormode = false;
 	opt->output_transbg = false;
 	opt->suppress_palette = false;
 }
 
+// (引数の) 文字列から ColorMode を返す。
+// エラーなら COLOR_MODE_NONE を返す。
+ColorMode
+image_parse_color(const char *arg)
+{
+	ColorMode color = COLOR_MODE_NONE;
+	int num = -1;
+
+	// 数字とその前とに分ける。正規表現でいうと /(\w*)(\d*)/ みたいな感じ。
+	size_t dp = strcspn(arg, "0123456789");
+	char str[dp + 1];
+	strncpy(str, arg, dp);
+	str[dp] = '\0';
+	if (arg[dp] != '\0') {
+		char *end;
+		num = stou32def(&arg[dp], -1, &end);
+		if (*end != '\0') {
+			goto abort;
+		}
+	}
+
+	if (str[0] == '\0') {
+		// 数字だけの場合。
+		switch (num) {
+		 case 2:	color = MAKE_COLOR_MODE_GRAY(2);		break;
+		 case 8:	color = COLOR_MODE_8_RGB;				break;
+		 case 16:	color = COLOR_MODE_16_VGA;				break;
+		 case 256:	color = MAKE_COLOR_MODE_ADAPTIVE(256);	break;
+		 default:
+			break;
+		}
+	} else {
+		// 文字から始まっている場合。
+#if defined(SIXELV)
+		if (strcmp(arg, "fixed256") == 0) {
+			color = COLOR_MODE_256_RGB332;
+		} else if (strcmp(arg, "xterm256") == 0) {
+			color = COLOR_MODE_256_XTERM;
+		} else if (strcmp(str, "adaptive") == 0) {
+			if (num < 0) {
+				color = MAKE_COLOR_MODE_ADAPTIVE(256);
+			} else if (8 <= num && num <= 256) {
+				color = MAKE_COLOR_MODE_ADAPTIVE(num);
+			}
+		} else
+#endif
+		if (strcmp(str, "gray") == 0 || strcmp(str, "grey") == 0) {
+			if (num < 0) {
+				color = MAKE_COLOR_MODE_GRAY(256);
+			} else if (2 <= num && num <= 256) {
+				color = MAKE_COLOR_MODE_GRAY(num);
+			}
+		}
+	}
+
+ abort:
+	return color;
+}
+
 // width_ x height_ で形式が format_ の image を作成する。
 // (バッファは未初期化)
 struct image *
@@ -281,6 +372,7 @@ image_get_preferred_size(
 #define LOADERMAP_gif	(1U << IMAGE_LOADER_GIF)
 #define LOADERMAP_jpeg	(1U << IMAGE_LOADER_JPEG)
 #define LOADERMAP_png	(1U << IMAGE_LOADER_PNG)
+#define LOADERMAP_tiff	(1U << IMAGE_LOADER_TIFF)
 #define LOADERMAP_webp	(1U << IMAGE_LOADER_WEBP)
 #define LOADERMAP_stb	((1U << IMAGE_LOADER_BMP)	| \
 						 (1U << IMAGE_LOADER_GIF)	| \
@@ -311,6 +403,9 @@ static const struct {
 #if defined(USE_GIFLIB)
 	ENTRY(gif, giflib, LOADERMAP_gif),
 #endif
+#if defined(USE_LIBTIFF) && defined(SIXELV)
+	ENTRY(tiff, libtiff, LOADERMAP_tiff),
+#endif
 #if defined(USE_STB_IMAGE)
 	ENTRY(stb, stb_image, LOADERMAP_stb),
 #endif
@@ -363,14 +458,14 @@ image_get_loaderinfo(void)
 }
 
 // pstream の画像形式を判定する。
-// 判定出来れば、image_read() に渡すための非負のインデックスを返す。
+// 判定出来れば非負のローダ種別を返す。これは image_read() に渡すのに使う。
 // 判定出来なければ -1 を返す。
 // ここでは Blurhash は扱わない。
 int
 image_match(struct pstream *ps, const struct diag *diag)
 {
 	FILE *fp;
-	int idx = -1;
+	int type = -1;
 
 	fp = pstream_open_for_peek(ps);
 	if (fp == NULL) {
@@ -389,7 +484,7 @@ image_match(struct pstream *ps, const struct diag *diag)
 			loader[i].name, (ok ? "matched" : "no"));
 		fseek(fp, 0, SEEK_SET);
 		if (ok) {
-			idx = i;
+			type = i;
 			goto done;
 		}
 	}
@@ -397,16 +492,16 @@ image_match(struct pstream *ps, const struct diag *diag)
 
  done:
 	fclose(fp);
-	return idx;
+	return type;
 }
 
 // pstream から画像を読み込んで image を作成して返す。
-// idx は image_match() で返された非負の値を指定すること。
+// type は image_match() で返されたローダ種別。
 // axis, width, height はリサイズ用のヒントで、これを使うかどうかは
 // 画像ローダによる (今の所 jpeg のみ)。
 // デコードに失敗すると NULL を返す。
 struct image *
-image_read(struct pstream *ps, int idx, const image_read_hint *hint,
+image_read(struct pstream *ps, int type, const image_read_hint *hint,
 	const struct diag *diag)
 {
 	FILE *fp;
@@ -416,7 +511,7 @@ image_read(struct pstream *ps, int idx, const image_read_hint *hint,
 		Debug(diag, "%s: pstream_open_for_read() failed", __func__);
 		return NULL;
 	}
-	struct image *img = loader[idx].read(fp, hint, diag);
+	struct image *img = loader[type].read(fp, hint, diag);
 	fclose(fp);
 
 	return img;
@@ -463,16 +558,19 @@ image_convert_to16(struct image *img)
 
 // src 画像を (dst_width, dst_height) にリサイズしながら同時に
 // colormode に減色した新しい image を作成して返す。
+// 適応パレット(COLOR_MODE_ADAPTIVE) なら、デバッグ表示用に
+// src->palette_count に色数を書き戻す。
 struct image *
 image_reduct(
-	const struct image *src,	// 元画像
+	struct image *src,			// 元画像
 	uint dst_width,				// リサイズ後の幅
 	uint dst_height,			// リサイズ後の高さ
-	const image_opt *opt,		// パラメータ
+	const struct image_opt *opt,	// パラメータ
 	const struct diag *diag)
 {
 	struct image *dst;
 	image_reductor_handle irbuf, *ir;
+	bool ok = false;
 
 	ir = &irbuf;
 	memset(ir, 0, sizeof(*ir));
@@ -484,52 +582,68 @@ image_reduct(
 	}
 	dst->has_alpha = src->has_alpha;
 
+	ir->dstimg = dst;
+	ir->srcimg = src;
+
 	// 減色モードからパレットオペレーションを用意。
-	switch (opt->color & COLOR_FMT_MASK) {
-	 case COLOR_FMT_GRAY:
+	switch (GET_COLOR_MODE(opt->color)) {
+	 case COLOR_MODE_GRAY:
 	 {
 		uint graycount = GET_COLOR_COUNT(opt->color);
-		ir->palette_buf = image_alloc_gray_palette(graycount);
-		if (ir->palette_buf == NULL) {
+		dst->palette_buf = image_alloc_gray_palette(graycount);
+		if (dst->palette_buf == NULL) {
 			goto abort;
 		}
-		ir->palette = ir->palette_buf;
-		ir->palette_count = graycount;
+		dst->palette = dst->palette_buf;
+		dst->palette_count = graycount;
 		ir->finder = finder_gray;
 		ir->is_gray = true;
 		break;
 	 }
 
-	 case COLOR_FMT_8_RGB:
+	 case COLOR_MODE_8_RGB:
 		ir->finder  = finder_fixed8;
-		ir->palette = palette_fixed8;
-		ir->palette_count = 8;
+		dst->palette = palette_fixed8;
+		dst->palette_count = 8;
 		break;
 
-	 case COLOR_FMT_16_VGA:
+	 case COLOR_MODE_16_VGA:
 		ir->finder  = finder_vga16;
-		ir->palette = palette_vga16;
-		ir->palette_count = 16;
+		dst->palette = palette_vga16;
+		dst->palette_count = 16;
 		break;
 
-	 case COLOR_FMT_256_RGB332:
-		ir->palette_buf = image_alloc_fixed256_palette();
-		if (ir->palette_buf == NULL) {
+	 case COLOR_MODE_ADAPTIVE:
+	 {
+		uint palcount = GET_COLOR_COUNT(opt->color);
+		dst->palette_buf = calloc(palcount, sizeof(ColorRGB));
+		if (dst->palette_buf == NULL) {
 			goto abort;
 		}
-		ir->palette = ir->palette_buf;
-		ir->palette_count = 256;
+		dst->palette = dst->palette_buf;
+		dst->palette_count = palcount;
+		ir->finder = finder_adaptive;
+		break;
+	 }
+
+#if defined(SIXELV)
+	 case COLOR_MODE_256_RGB332:
+		dst->palette_buf = image_alloc_fixed256_palette();
+		if (dst->palette_buf == NULL) {
+			goto abort;
+		}
+		dst->palette = dst->palette_buf;
+		dst->palette_count = 256;
 		ir->finder = finder_fixed256;
 		break;
 
-#if defined(SIXELV)
-	 case COLOR_FMT_256_XTERM:
-		ir->palette_buf = image_alloc_xterm256_palette();
-		if (ir->palette_buf == NULL) {
+	 case COLOR_MODE_256_XTERM:
+		dst->palette_buf = image_alloc_xterm256_palette();
+		if (dst->palette_buf == NULL) {
 			goto abort;
 		}
-		ir->palette = ir->palette_buf;
-		ir->palette_count = 256;
+		dst->palette = dst->palette_buf;
+		dst->palette_count = 256;
 		ir->finder = finder_xterm256;
 		break;
 #endif
@@ -539,34 +653,40 @@ image_reduct(
 		goto abort;
 	}
 
+#if defined(IMAGE_PROFILE)
+	memset(&y_count, 0, sizeof(y_count));
+#endif
+
 #if defined(SIXELV)
 	if (opt->method == REDUCT_SIMPLE) {
-		image_reduct_simple(ir, dst, src, diag);
+		ok = image_reduct_simple(ir, opt, diag);
 
 	} else if (opt->method != REDUCT_HIGH_QUALITY) {
 		Debug(diag, "%s: Unknown method %u", __func__, opt->method);
-		goto abort;
+		ok = false;
 
 	} else
 #endif
 	{
-		if (image_reduct_highquality(ir, dst, src, opt, diag) == false) {
-			goto abort;
-		}
+		ok = image_reduct_highquality(ir, opt, diag);
 	}
 
-	// 成功したので、使ったパレットを image にコピー。
-	// 動的に確保したやつはそのまま所有権を移す感じ。
-	dst->palette       = ir->palette;
-	dst->palette_count = ir->palette_count;
-	dst->palette_buf   = ir->palette_buf;
-
-	return dst;
+#if defined(IMAGE_PROFILE)
+	printf("[lo, hi )  count\n");
+	for (uint i = 0; i < 8; i++) {
+		printf("%3u, %3u = %u\n", ir->y_lo[i], ir->y_hi[i], y_count[i]);
+	}
+#endif
 
  abort:
-	free(ir->palette_buf);
-	image_free(dst);
-	return NULL;
+#if defined(SIXELV)
+	free(ir->colorhash);
+#endif
+	if (!ok) {
+		image_free(dst);
+		dst = NULL;
+	}
+	return dst;
 }
 
 
@@ -615,10 +735,12 @@ rational_add(Rational *sr, const Rational *x)
 
 #if defined(SIXELV)
 // 単純間引き
-static void
+static bool
 image_reduct_simple(image_reductor_handle *ir,
-	struct image *dstimg, const struct image *srcimg, const struct diag *diag)
+	const struct image_opt *opt, const struct diag *diag)
 {
+	struct image *dstimg = ir->dstimg;
+	const struct image *srcimg = ir->srcimg;
 	uint16 *d = (uint16 *)dstimg->buf;
 	const uint16 *src = (const uint16 *)srcimg->buf;
 	uint dstwidth  = dstimg->width;
@@ -628,6 +750,13 @@ image_reduct_simple(image_reductor_handle *ir,
 	Rational ystep;
 	Rational xstep;
 
+	// 適応パレットならここでパレットを作成。
+	if (GET_COLOR_MODE(opt->color) == COLOR_MODE_ADAPTIVE) {
+		if (image_calc_adaptive_palette(ir) == false) {
+			return false;
+		}
+	}
+
 	// 水平、垂直方向ともスキップサンプリング。
 
 	rational_init(&ry, 0, 0, dstheight);
@@ -640,28 +769,26 @@ image_reduct_simple(image_reductor_handle *ir,
 		rx.N = 0;
 		const uint16 *s0 = &src[ry.I * srcimg->width];
 		for (uint x = 0; x < dstwidth; x++) {
-			const uint16 *s = s0 + rx.I;
-
-			ColorRGBint32 c;
+			ColorRGBint32 col;
 			int a;
-			uint16 v = *s++;
-			a   =  (v >> 15);
-			c.r = ((v >> 10) & 0x1f) << 3;
-			c.g = ((v >>  5) & 0x1f) << 3;
-			c.b = ( v        & 0x1f) << 3;
+			uint16 v = s0[rx.I];
+			a     =  (v >> 15);
+			col.r = ((v >> 10) & 0x1f) << 3;
+			col.g = ((v >>  5) & 0x1f) << 3;
+			col.b = ( v        & 0x1f) << 3;
 
-			if (ir->gain != 256) {
-				c.r = (uint32)c.r * ir->gain / 256;
-				c.g = (uint32)c.g * ir->gain / 256;
-				c.b = (uint32)c.b * ir->gain / 256;
+			if (ir->gain >= 0) {
+				col.r = saturate_uint8((uint32)col.r * ir->gain / 256);
+				col.g = saturate_uint8((uint32)col.g * ir->gain / 256);
+				col.b = saturate_uint8((uint32)col.b * ir->gain / 256);
 			}
 			if (ir->is_gray) {
-				colorcvt_gray(&c);
+				colorcvt_gray(&col);
 			}
 			ColorRGB c8;
-			c8.r = saturate_uint8(c.r);
-			c8.g = saturate_uint8(c.g);
-			c8.b = saturate_uint8(c.b);
+			c8.r = col.r;
+			c8.g = col.g;
+			c8.b = col.b;
 			uint colorcode = ir->finder(ir, c8);
 			if (a) {
 				colorcode |= 0x8000;
@@ -672,15 +799,18 @@ image_reduct_simple(image_reductor_handle *ir,
 		}
 		rational_add(&ry, &ystep);
 	}
+
+	return true;
 }
 #endif
 
 // 二次元誤差分散法を使用して、出来る限り高品質に変換する。
 static bool
 image_reduct_highquality(image_reductor_handle *ir,
-	struct image *dstimg, const struct image *srcimg, const image_opt *opt,
-	const struct diag *diag)
+	const struct image_opt *opt, const struct diag *diag)
 {
+	struct image *dstimg = ir->dstimg;
+	const struct image *srcimg = ir->srcimg;
 	uint16 *d = (uint16 *)dstimg->buf;
 	const uint16 *src = (const uint16 *)srcimg->buf;
 	uint dstwidth  = dstimg->width;
@@ -697,6 +827,13 @@ image_reduct_highquality(image_reductor_handle *ir,
 	assert(opt->diffuse == DIFFUSE_SFL);
 #endif
 
+	// 適応パレットならここでパレットを作成。
+	if (GET_COLOR_MODE(opt->color) == COLOR_MODE_ADAPTIVE) {
+		if (image_calc_adaptive_palette(ir) == false) {
+			return false;
+		}
+	}
+
 	// 水平、垂直ともピクセルを平均。
 	// 真に高品質にするには補間法を適用するべきだがそこまではしない。
 
@@ -759,7 +896,7 @@ image_reduct_highquality(image_reductor_handle *ir,
 			col.g /= area;
 			col.b /= area;
 
-			if (ir->gain != 256) {
+			if (ir->gain >= 0) {
 				col.r = (uint32)col.r * ir->gain / 256;
 				col.g = (uint32)col.g * ir->gain / 256;
 				col.b = (uint32)col.b * ir->gain / 256;
@@ -798,9 +935,9 @@ image_reduct_highquality(image_reductor_handle *ir,
 			}
 			*d++ = v;
 
-			col.r -= ir->palette[colorcode].r;
-			col.g -= ir->palette[colorcode].g;
-			col.b -= ir->palette[colorcode].b;
+			col.r -= dstimg->palette[colorcode].r;
+			col.g -= dstimg->palette[colorcode].g;
+			col.b -= dstimg->palette[colorcode].b;
 
 			// ランダムノイズを加える。
 			if (0) {
@@ -834,12 +971,12 @@ image_reduct_highquality(image_reductor_handle *ir,
 				break;
 			 case DIFFUSE_ATKINSON:
 				// Atkinson
-				set_err(errbuf[0], x + 1, &col, 32);
-				set_err(errbuf[0], x + 2, &col, 32);
-				set_err(errbuf[1], x - 1, &col, 32);
-				set_err(errbuf[1], x,     &col, 32);
-				set_err(errbuf[1], x + 1, &col, 32);
-				set_err(errbuf[2], x,     &col, 32);
+				set_err_asr(errbuf[0], x + 1, &col, 3);	// 32
+				set_err_asr(errbuf[0], x + 2, &col, 3);	// 32
+				set_err_asr(errbuf[1], x - 1, &col, 3);	// 32
+				set_err_asr(errbuf[1], x,     &col, 3);	// 32
+				set_err_asr(errbuf[1], x + 1, &col, 3);	// 32
+				set_err_asr(errbuf[2], x,     &col, 3);	// 32
 				break;
 			 case DIFFUSE_JAJUNI:
 				// Jarvis, Judice, Ninke
@@ -873,13 +1010,13 @@ image_reduct_highquality(image_reductor_handle *ir,
 				break;
 			 case DIFFUSE_BURKES:
 				// Burkes
-				set_err(errbuf[0], x + 1, &col, 64);
-				set_err(errbuf[0], x + 2, &col, 32);
-				set_err(errbuf[1], x - 2, &col, 16);
-				set_err(errbuf[1], x - 1, &col, 32);
-				set_err(errbuf[1], x,     &col, 64);
-				set_err(errbuf[1], x + 1, &col, 32);
-				set_err(errbuf[1], x + 2, &col, 16);
+				set_err_asr(errbuf[0], x + 1, &col, 2);	// 64
+				set_err_asr(errbuf[0], x + 2, &col, 3);	// 32
+				set_err_asr(errbuf[1], x - 2, &col, 4);	// 16
+				set_err_asr(errbuf[1], x - 1, &col, 3);	// 32
+				set_err_asr(errbuf[1], x,     &col, 2);	// 64
+				set_err_asr(errbuf[1], x + 1, &col, 3);	// 32
+				set_err_asr(errbuf[1], x + 2, &col, 4);	// 16
 				break;
 			 case DIFFUSE_2:
 				// (x+1,y), (x,y+1)
@@ -984,11 +1121,11 @@ image_alloc_gray_palette(uint count)
 	return pal;
 }
 
-// 256 段階グレースケールになっている c からパレット番号を返す。
+// count 段階グレースケールになっている c からパレット番号を返す。
 static uint
 finder_gray(image_reductor_handle *ir, ColorRGB c)
 {
-	uint count = ir->palette_count;
+	uint count = ir->dstimg->palette_count;
 
 	int I = (((uint)c.r) * (count - 1) + (255 / count)) / 255;
 	if (I >= count) {
@@ -998,12 +1135,10 @@ finder_gray(image_reductor_handle *ir, ColorRGB c)
 }
 
 // c をグレー (NTSC 輝度) に変換する。
-// I = (R * 0.299) + (G * 0.587) + (B * 0.114)
 static void
 colorcvt_gray(ColorRGBint32 *c)
 {
-	// 元々この係数の厳密性はそんなにないので高速化のため雑に近似。
-	int I = (c->r * 5 + c->g * 9 + c->b * 2) / 16;
+	int I = RGBToY(c->r, c->g, c->b);
 	c->r = I;
 	c->g = I;
 	c->b = I;
@@ -1109,6 +1244,8 @@ finder_vga16(image_reductor_handle *ir, ColorRGB c)
 	}
 }
 
+#if defined(SIXELV)
+
 // R3,G3,B2 の固定 256 色パレットを作成して返す。
 static ColorRGB *
 image_alloc_fixed256_palette(void)
@@ -1138,8 +1275,6 @@ finder_fixed256(image_reductor_handle *ir, ColorRGB c)
 	return (R << 5) | (G << 2) | B;
 }
 
-#if defined(SIXELV)
-
 // xterm 互換の固定 256 色パレットを作成して返す。
 static ColorRGB *
 image_alloc_xterm256_palette(void)
@@ -1170,7 +1305,10 @@ image_alloc_xterm256_palette(void)
 	// グレー24色。
 	for (i = 0; i < 24; i++) {
 		ColorRGB c;
-		c.r = c.g = c.b = 8 + i * 10;
+		uint8 I = 8 + i * 10;
+		c.r = I;
+		c.g = I;
+		c.b = I;
 		pal[i + 16 + 216] = c;
 	}
 
@@ -1206,6 +1344,443 @@ finder_xterm256(image_reductor_handle *ir, ColorRGB c)
 		+ finder_xterm256_channel(c.b) * 1;
 }
 
+#endif // SIXELV
+
+//
+// 適応 256 色パレット。
+//
+
+struct octree {
+	uint32 count;	// ピクセル数
+	uint32 r;		// R 合計
+	uint32 g;		// G 合計
+	uint32 b;		// B 合計
+	struct octree *children; // [8]
+};
+
+static const uint16 tobits[] = {
+	0x0000,	// 0b000000000000000
+	0x1000,	// 0b001000000000000
+	0x0200,	// 0b000001000000000
+	0x1200,	// 0b001001000000000
+	0x0040,	// 0b000000001000000
+	0x1040,	// 0b001000001000000
+	0x0240,	// 0b000001001000000
+	0x1240,	// 0b001001001000000
+	0x0008,	// 0b000000000001000
+	0x1008,	// 0b001000000001000
+	0x0208,	// 0b000001000001000
+	0x1208,	// 0b001001000001000
+	0x0048,	// 0b000000001001000
+	0x1048,	// 0b001000001001000
+	0x0248,	// 0b000001001001000
+	0x1248,	// 0b001001001001000
+	0x0001,	// 0b000000000000001
+	0x1001,	// 0b001000000000001
+	0x0201,	// 0b000001000000001
+	0x1201,	// 0b001001000000001
+	0x0041,	// 0b000000001000001
+	0x1041,	// 0b001000001000001
+	0x0241,	// 0b000001001000001
+	0x1241,	// 0b001001001000001
+	0x0009,	// 0b000000000001001
+	0x1009,	// 0b001000000001001
+	0x0209,	// 0b000001000001001
+	0x1209,	// 0b001001000001001
+	0x0049,	// 0b000000001001001
+	0x1049,	// 0b001000001001001
+	0x0249,	// 0b000001001001001
+	0x1249,	// 0b001001001001001
+};
+
+// octree に色を配置する。
+// bits は c の R,G,B (の上の 5 ビット) を左右逆にして R,G,B シャッフルした
+// ビット列で、これで下から3ビットずつ取り出すとそのまま octree の各階層の
+// インデックスになる。
+//  14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
+// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
+// |R3|G3|B3|R4|G4|B4|R5|G5|B5|R6|G6|B6|R7|G7|B7|
+// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
+static bool
+octree_set(struct octree *node, uint32 bits, ColorRGB c, uint32 count)
+{
+	for (uint lv = 0; lv < 5; lv++) {
+		// node->count は自ノード以下のピクセル数なので、途中にもすべて加算。
+		node->count += count;
+
+		if (__predict_false(node->children == NULL)) {
+			node->children = calloc(8, sizeof(struct octree));
+			if (__predict_false(node->children == NULL)) {
+				return false;
+			}
+		}
+
+		node = &node->children[(bits & 7)];
+		bits >>= 3;
+	}
+
+	// リーフにデータを置く。
+	// ここは色ごとに一度ずつしか呼ばないので代入でいい。
+	node->count = count;
+	node->r = c.r * count;
+	node->g = c.g * count;
+	node->b = c.b * count;
+
+	return true;
+}
+
+// node 以下で count が最小である、リーフ直上のノードを返す。
+// *min は in/out パラメータで、初期 min を渡す。
+// min より最小のものが見付かれば *min を更新してそのノードを返す。
+// 見付からなければ *min を更新せず NULL を返す。
+static struct octree *
+octree_find_minnode(struct octree *node, uint32 *min)
+{
+	if (__predict_false(node->children == NULL)) {
+		// リーフ。ここには来ないはず。
+		return NULL;
+	}
+
+	// 子ノードのいずれかが孫を持つか。
+	bool has_grandchild =
+		(node->children[0].children != NULL) |
+		(node->children[1].children != NULL) |
+		(node->children[2].children != NULL) |
+		(node->children[3].children != NULL) |
+		(node->children[4].children != NULL) |
+		(node->children[5].children != NULL) |
+		(node->children[6].children != NULL) |
+		(node->children[7].children != NULL);
+
+	if (has_grandchild) {
+		// 自分はまだ中間ノード。
+		struct octree *minnode = NULL;
+		for (uint i = 0; i < 8; i++) {
+			struct octree *n = octree_find_minnode(&node->children[i], min);
+			if (n) {
+				minnode = n;
+			}
+		}
+		return minnode;
+	} else {
+		// 自分の子がリーフのノード。
+		if (node->count < *min) {
+			*min = node->count;
+			return node;
+		} else {
+			return NULL;
+		}
+	}
+}
+
+// このノードのリーフをマージする。リーフ直上のノードで行うこと。
+// 戻り値はマージによって増減したリーフ数。子リーフを1〜8個減らすが、
+// 自分が新たにリーフになって 1 増えるので、都合 0 以下になる。
+static int
+octree_merge_leaves(struct octree *node)
+{
+	uint32 r = 0;
+	uint32 g = 0;
+	uint32 b = 0;
+	int ndiff = 1;
+
+	for (uint i = 0; i < 8; i++) {
+		struct octree *child = &node->children[i];
+		if (child->count != 0) {
+			ndiff--;
+			r += child->r;
+			g += child->g;
+			b += child->b;
+		}
+	}
+	// count は計算済み。
+	node->r = r;
+	node->g = g;
+	node->b = b;
+	free(node->children);
+	node->children = NULL;
+
+	return ndiff;
+}
+
+// Y (輝度)でソートする。といいつつ Y は .a のところを使っている。
+static int
+cmp_y(const void *a1, const void *a2)
+{
+	const ColorRGB *c1 = a1;
+	const ColorRGB *c2 = a2;
+
+	return c1->a - c2->a;
+}
+
+// node 以下のリーフをパレットに登録していく。
+// n は次のパレット番号。
+// 戻り値も次のパレット番号。
+static uint
+octree_make_palette(ColorRGB *pal, uint n, const struct octree *node)
+{
+	if (node->children) {
+		for (uint i = 0; i < 8; i++) {
+			n = octree_make_palette(pal, n, &node->children[i]);
+		}
+	} else {
+		if (node->count != 0) {
+			uint32 r = node->r / node->count;
+			uint32 g = node->g / node->count;
+			uint32 b = node->b / node->count;
+			// 輝度を計算して A のところに入れておく。
+			// (パレットの .A はパレットとしては使っていない)
+			uint32 y = RGBToY(r, g, b);
+
+			pal[n].r = r;
+			pal[n].g = g;
+			pal[n].b = b;
+			pal[n].a = y;
+			n++;
+		}
+	}
+	return n;
+}
+
+// パレットから c に最も近い色のパレット番号を返す。
+static uint
+finder_linear(image_reductor_handle *ir, ColorRGB c)
+{
+	uint32 mindist = (uint32)-1;
+	uint minidx = 0;
+
+	uint32 yh = RGBToY(c.r, c.g, c.b) >> 5;
+
+#if defined(IMAGE_PROFILE)
+	y_count[yh]++;
+#endif
+	for (uint i = ir->y_lo[yh]; i < ir->y_hi[yh]; i++) {
+		const ColorRGB *pal = &ir->dstimg->palette[i];
+		int32 dr = c.r - pal->r;
+		int32 dg = c.g - pal->g;
+		int32 db = c.b - pal->b;
+		int32 dist = (dr * dr) + (dg * dg) + (db * db);
+
+		if (dist < mindist) {
+			minidx = i;
+			if (__predict_false(dist < 8)) {
+				break;
+			}
+			mindist = dist;
+		}
+	}
+	return minidx;
+}
+
+// node の子を解放する。node 自身はこの親が解放すること。
+static void
+octree_free(struct octree *node)
+{
+	if (node->children) {
+		for (uint i = 0; i < 8; i++) {
+			octree_free(&node->children[i]);
+		}
+		free(node->children);
+		node->children = NULL;
+	}
+}
+
+// srcimg から適応パレットを作成。
+static bool
+image_calc_adaptive_palette(image_reductor_handle *ir)
+{
+	struct image *dstimg = ir->dstimg;
+	struct image *srcimg = ir->srcimg;
+	const uint16 *src = (const uint16 *)srcimg->buf;
+	uint palette_count;
+	struct octree root;
+	bool rv = false;
+#if defined(IMAGE_PROFILE)
+	struct timespec colormap_start, colormap_end;
+	struct timespec octree_start, octree_end;
+	struct timespec merge_start, merge_end;
+	struct timespec make_start, make_end;
+#endif
+
+	// この直後で使う colormap は uint16 * 32768。
+	// 一方この関数を終えて reduct 中に使う ir->colorhash も uint16 * 32768 で
+	// 両者は使用期間がかぶらないので、一度確保したのを使い回す。
+	const uint32 capacity = 32768;
+	ir->colorhash = calloc(capacity, sizeof(uint16));
+	if (__predict_false(ir->colorhash == NULL)) {
+		return false;
+	}
+	uint16 *colormap = ir->colorhash;
+
+	// src 画像に使われている色を全部取り出す。
+	// この時点で R,G,B 各色5ビットで足切りされていて、合計15ビットしか
+	// ないので、配列の添字にしてダイレクトアクセスする。
+	PROF(colormap_start);
+	const uint16 *send = src + srcimg->width * srcimg->height;
+	for (const uint16 *s = src; s < send; ) {
+		uint16 n = *s++;
+		n &= 0x7fff;
+		if (__predict_false(ir->gain >= 0)) {
+			uint32 r5 = ((n >> 10) & 0x1f) * ir->gain / 256;
+			uint32 g5 = ((n >>  5) & 0x1f) * ir->gain / 256;
+			uint32 b5 = ( n        & 0x1f) * ir->gain / 256;
+			if (__predict_false(r5 > 31)) r5 = 31;
+			if (__predict_false(g5 > 31)) g5 = 31;
+			if (__predict_false(b5 > 31)) b5 = 31;
+			n = (r5 << 10) | (g5 << 5) | b5;
+		}
+		// ピクセル数の計数は 0xffff で頭打ちにしておく。
+		// 65536 ピクセル以上の色が 256 色以上ある時には困るかも知れないが。
+		uint16 count = colormap[n];
+		uint16 count1 = count + 1;
+		if (__predict_true(count1 != 0)) {
+			colormap[n] = count1;
+		}
+	}
+	PROF(colormap_end);
+
+	// octree に配置。
+	palette_count = 0;
+	PROF(octree_start);
+	memset(&root, 0, sizeof(root));
+	for (uint i = 0; i < capacity; i++) {
+		uint32 count = colormap[i];
+		if (__predict_true(count == 0)) {
+			continue;
+		}
+		uint32 r5 = (i >> 10) & 0x1f;
+		uint32 g5 = (i >>  5) & 0x1f;
+		uint32 b5 = (i      ) & 0x1f;
+		uint32 bits = (tobits[r5] << 2) | (tobits[g5] << 1) | tobits[b5];
+		ColorRGB c;
+		c.r = (r5 << 3);
+		c.g = (g5 << 3);
+		c.b = (b5 << 3);
+		if (__predict_false(octree_set(&root, bits, c, count) == false)) {
+			goto abort;
+		}
+		palette_count++;
+	}
+	PROF(octree_end);
+	srcimg->palette_count = palette_count;
+
+	if (0) {
+		for (uint i = 0; i < 8; i++) {
+			printf("[%u] %u\n", i, root.children[i].count);
+			if (root.children) {
+				for (uint j = 0; j < 8; j++) {
+					struct octree *n1 = &root.children[j];
+					printf(" [%u,%u] %u\n", i, j, n1->count);
+					if (n1->children) {
+						for (uint k = 0; k < 8; k++) {
+							struct octree *n2 = &n1->children[k];
+							printf("  [%u,%u,%u] %u\n", i, j, k, n2->count);
+						}
+					}
+				}
+			}
+		}
+	}
+
+	// 指定の色数以下になるまで少ない色をマージしていく。
+	PROF(merge_start);
+	uint dst_count = dstimg->palette_count;
+	while (palette_count > dst_count) {
+		//printf("palette_count=%u\n", palette_count);
+		uint32 min = -1;
+		struct octree *minnode = octree_find_minnode(&root, &min);
+		palette_count += octree_merge_leaves(minnode);
+	}
+	PROF(merge_end);
+	dstimg->palette_count = palette_count;
+
+	// パレットにセット。
+	PROF(make_start);
+	ColorRGB *dstpal = dstimg->palette_buf;
+	octree_make_palette(dstpal, 0, &root);
+	// パレットを Y (輝度) でソート。
+	qsort(dstpal, palette_count, sizeof(dstpal[0]), cmp_y);
+	// Y の上位 3 ビット(8通り)に対応する検索範囲を事前に調べておく。
+	// どう見ても明るい色なら暗い方と比較する必要ないよねというくらい。
+	// margin の値は適当。狭いと速いが最も近い色を拾えない可能性がある、
+	// 広いと色は近くなる可能性が上がるが遅くなる。
+	// ただしそもそも RGB のユークリッド距離なので細かい事は気にしない。
+	const int margin = 20;
+	for (uint y = 0; y < 8; y++) {
+		int lo = y * 0x20 - margin;
+		if (lo < 0)
+			lo = 0;
+		int hi = y * 0x20 + (0x20 - 1) + margin;
+		if (hi > 255)
+			hi = 255;
+
+		int i = 0;
+		int ylo;
+		int yhi;
+		for (; i < dstimg->palette_count - 1; i++) {
+			if (dstpal[i].a >= lo) {
+				break;
+			}
+		}
+		ylo = i;
+		for (; i < dstimg->palette_count; i++) {
+			if (dstpal[i].a > hi) {
+				break;
+			}
+		}
+		yhi = i;
+
+		// 0 個になった場合。
+		if (ylo == yhi) {
+			if (yhi == dstimg->palette_count) {
+				ylo--;
+			} else {
+				yhi++;
+			}
+		}
+		ir->y_lo[y] = ylo;
+		ir->y_hi[y] = yhi;
+
+	}
+	PROF(make_end);
+
+	PROF_RESULT("colormap",		colormap);
+	PROF_RESULT("octree_set",	octree);
+	PROF_RESULT("octree_merge",	merge);
+	PROF_RESULT("octree_make",	make);
+
+	// colorhash を本当はここで確保するが
+	// 使い終わった colormap とサイズが同じなのでありがたく使い回す。
+	// まだパレット引いてない印の (uint16)-1 で初期化する。
+	memset(ir->colorhash, 0xff, capacity * sizeof(uint16));
+
+	rv = true;
+ abort:
+	octree_free(&root);
+	return rv;
+}
+
+// 適応パレットから c に最も近いパレット番号を返す。
+static uint
+finder_adaptive(image_reductor_handle *ir, ColorRGB c)
+{
+	uint32 r5 = c.r >> 3;
+	uint32 g5 = c.g >> 3;
+	uint32 b5 = c.b >> 3;
+	uint32 n = r5 * 32 * 32 + g5 * 32 + b5;
+	uint16 cc = ir->colorhash[n];
+	if (__predict_false((int16)cc < 0)) {
+		// まだパレット番号が入ってなければここで引く。
+		c.r = r5 * 8 + 4;
+		c.g = g5 * 8 + 4;
+		c.b = b5 * 8 + 4;
+		cc = finder_linear(ir, c);
+		ir->colorhash[n] = cc;
+	}
+	return cc;
+}
+
+#if defined(SIXELV)
 
 //
 // enum のデバッグ表示用
@@ -1275,27 +1850,30 @@ reductordiffuse_tostr(ReductorDiffuse diffuse)
 	return buf;
 }
 
-// ColorFormat を文字列にする。
+// ColorMode を文字列にする。
 // (内部バッファを使う可能性があるため同時に2回呼ばないこと)
 const char *
-colorformat_tostr(ColorFormat color)
+colormode_tostr(ColorMode color)
 {
 	static const struct {
-		ColorFormat value;
+		ColorMode mode;
 		const char *name;
 	} table[] = {
-		{ COLOR_FMT_GRAY,		"Gray" },
-		{ COLOR_FMT_8_RGB,		"8(RGB)" },
-		{ COLOR_FMT_16_VGA,		"16(ANSI VGA)" },
-		{ COLOR_FMT_256_RGB332,	"256(RGB332)" },
-		{ COLOR_FMT_256_XTERM,	"256(xterm)" },
+		{ COLOR_MODE_GRAY,			"Gray" },
+		{ COLOR_MODE_8_RGB,			"8(RGB)" },
+		{ COLOR_MODE_16_VGA,		"16(ANSI VGA)" },
+		{ COLOR_MODE_256_RGB332,	"256(RGB332)" },
+		{ COLOR_MODE_256_XTERM,		"256(xterm)" },
+		{ COLOR_MODE_ADAPTIVE,		"Adaptive" },
 	};
 	static char buf[16];
-	uint type = (uint)color & COLOR_FMT_MASK;
+	uint colormode = GET_COLOR_MODE(color);
 
 	for (int i = 0; i < countof(table); i++) {
-		if (table[i].value == type) {
-			if (type == COLOR_FMT_GRAY) {
+		if (table[i].mode == colormode) {
+			if (colormode == COLOR_MODE_GRAY ||
+			    colormode == COLOR_MODE_ADAPTIVE)
+			{
 				uint num = GET_COLOR_COUNT(color);
 				snprintf(buf, sizeof(buf), "%s%u", table[i].name, num);
 				return buf;
