@@ -118,21 +118,27 @@ typedef enum {
 } ReductorDiffuse;
 
 // 色モードは下位8ビットが enum。
-// Gray、GrayMean では bit15-8 の 8ビットに「階調-1」(1-255) を格納する。
+// GRAY では bit15-8 の 8ビットに「階調-1」(=1-255) を格納する。
+// ADAPTIVE も同様に「色数-1」(=7-255) を格納する。
 typedef enum {
-	COLOR_FMT_GRAY,			// グレイスケール
-	COLOR_FMT_8_RGB,		// RGB 8色
-	COLOR_FMT_16_VGA,		// ANSI 16色 (VGA)
-	COLOR_FMT_256_RGB332,	// 固定 256 色 (MSX SCREEN8 互換)
-	COLOR_FMT_256_XTERM,	// 固定 256 色 (xterm 互換)
+	COLOR_MODE_NONE,
+	COLOR_MODE_GRAY,			// グレイスケール
+	COLOR_MODE_8_RGB,			// RGB 8色
+	COLOR_MODE_16_VGA,			// ANSI 16色 (VGA)
+	COLOR_MODE_256_RGB332,		// 固定 256 色 (MSX SCREEN8 互換)
+	COLOR_MODE_256_XTERM,		// 固定 256 色 (xterm 互換)
+	COLOR_MODE_ADAPTIVE,		// 適応パレット (8-256 色)
 
 	// 下位8ビットが色モード
-	COLOR_FMT_MASK = 0xff,
-} ColorFormat;
+	COLOR_MODE_MASK = 0xff,
+} ColorMode;
 
-#define MAKE_COLOR_FMT_GRAY(n)	\
-	(COLOR_FMT_GRAY | (((uint)(n) - 1) << 8))
+#define MAKE_COLOR_MODE_GRAY(n)	\
+	(COLOR_MODE_GRAY | (((uint)(n) - 1) << 8))
+#define MAKE_COLOR_MODE_ADAPTIVE(n)	\
+	(COLOR_MODE_ADAPTIVE | (((uint)(n) - 1) << 8))
 
+#define GET_COLOR_MODE(fmt)		((fmt) & COLOR_MODE_MASK)
 #define GET_COLOR_COUNT(fmt)	((((uint)(fmt)) >> 8) + 1)
 
 #if BYTE_ORDER == LITTLE_ENDIAN
@@ -167,6 +173,7 @@ enum {
 	IMAGE_LOADER_JPEG,
 	IMAGE_LOADER_PNG,
 	IMAGE_LOADER_PNM,
+	IMAGE_LOADER_TIFF,
 	IMAGE_LOADER_WEBP,
 	IMAGE_LOADER_MAX,
 };
@@ -178,6 +185,7 @@ enum {
 	"jpeg",	\
 	"png",	\
 	"pnm",	\
+	"tiff",	\
 	"webp"
 
 struct image
@@ -200,24 +208,29 @@ struct image
 	const ColorRGB *palette;
 
 	// パレット数。インデックスカラーでない場合は参照しないこと。
+	// ただし、適応パレットに変換する際の入力画像ならデバッグ表示用に
+	// ここに色数を入れる。
 	uint palette_count;
 
 	// 動的に作成したパレットはここで所有する。
 	ColorRGB *palette_buf;
 };
 
-// image_read() に対するサイズヒント。
+// image_read() に対するヒント。
 typedef struct image_read_hint_ {
 	ResizeAxis axis;
 	uint width;
 	uint height;
+
+	// 複数枚ある場合のページ(フレーム)番号。0 から始まる。
+	uint page;
 } image_read_hint;
 
-typedef struct image_opt_ {
+struct image_opt {
 	// 減色
 	ReductorMethod method;
 	ReductorDiffuse diffuse;
-	ColorFormat color;
+	ColorMode color;
 
 	// 誤差の減衰率(?)。0 .. 256 で指定する。
 	// 256以上は 256 と同じ効果となる。
@@ -225,16 +238,18 @@ typedef struct image_opt_ {
 	uint cdm;
 
 	// 出力ゲイン。0 .. 512 で指定する。256 が 1.0倍。
-	uint gain;
+	// 負数なら適用しない (1.0 倍と同じ)。
+	int gain;
 
 	// SIXEL 出力
 	bool output_ormode;
 	bool output_transbg;
 	bool suppress_palette;
-} image_opt;
+};
 
 // image.c
-extern void image_opt_init(image_opt *);
+extern void image_opt_init(struct image_opt *);
+extern ColorMode image_parse_color(const char *);
 extern int  image_match(struct pstream *, const struct diag *);
 extern struct image *image_read(struct pstream *, int,
 	const image_read_hint *, const struct diag *);
@@ -245,19 +260,19 @@ extern void image_get_preferred_size(uint, uint, ResizeAxis,
 	uint, uint, uint *, uint *);
 extern char **image_get_loaderinfo(void);
 extern void image_convert_to16(struct image *);
-extern struct image *image_reduct(const struct image *, uint, uint,
-	const image_opt *, const struct diag *);
+extern struct image *image_reduct(struct image *, uint, uint,
+	const struct image_opt *, const struct diag *);
 
 extern const char *resizeaxis_tostr(ResizeAxis);
 extern const char *reductordiffuse_tostr(ReductorDiffuse);
-extern const char *colorformat_tostr(ColorFormat);
+extern const char *colormode_tostr(ColorMode);
 
 // image_blurhash.c
 extern struct image *image_blurhash_read(FILE *, int, int, const struct diag *);
 
 // image_sixel.c
 extern void image_sixel_abort(FILE *);
-extern bool image_sixel_write(FILE *, const struct image *, const image_opt *,
-	const struct diag *);
+extern bool image_sixel_write(FILE *, const struct image *,
+	const struct image_opt *, const struct diag *);
 
 #endif // !sayaka_image_h
