@@ -32,6 +32,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <string.h>
+#include <time.h>
 #include <unistd.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -50,7 +51,7 @@ struct net {
 		const struct net_opt *);
 	int (*f_read)(struct net *, void *, int);
 	int (*f_write)(struct net *, const void *, int);
-	void (*f_shutdown)(struct net *);
+	void (*f_shutdown_half)(struct net *);
 	void (*f_close)(struct net *);
 
 	// f_connect() が確保したリソースを解放する。
@@ -78,7 +79,7 @@ static int  sock_connect(struct net *, const char *, const char *,
 	const struct net_opt *);
 static int  sock_read(struct net *, void *, int);
 static int  sock_write(struct net *, const void *, int);
-static void sock_shutdown(struct net *);
+static void sock_shutdown_half(struct net *);
 static void sock_close(struct net *);
 #if defined(HAVE_OPENSSL)
 static void tls_cleanup(struct net *);
@@ -86,7 +87,7 @@ static int  tls_connect(struct net *, const char *, const char *,
 	const struct net_opt *);
 static int  tls_read(struct net *, void *, int);
 static int  tls_write(struct net *, const void *, int);
-static void tls_shutdown(struct net *);
+static void tls_shutdown_half(struct net *);
 static void tls_close(struct net *);
 #endif
 static int  socket_connect(const char *, const char *, const struct net_opt *);
@@ -334,7 +335,7 @@ net_connect(struct net *net,
 		net->f_connect = tls_connect;
 		net->f_read    = tls_read;
 		net->f_write   = tls_write;
-		net->f_shutdown= tls_shutdown;
+		net->f_shutdown_half = tls_shutdown_half;
 		net->f_close   = tls_close;
 		net->f_cleanup = tls_cleanup;
 #else
@@ -345,7 +346,7 @@ net_connect(struct net *net,
 		net->f_connect = sock_connect;
 		net->f_read    = sock_read;
 		net->f_write   = sock_write;
-		net->f_shutdown= sock_shutdown;
+		net->f_shutdown_half = sock_shutdown_half;
 		net->f_close   = sock_close;
 		net->f_cleanup = sock_cleanup;
 	}
@@ -356,6 +357,8 @@ net_connect(struct net *net,
 }
 
 // 1行受信して返す。
+// EOF に到達したらそこまでに受信した文字列を返す (行頭で EOF なら "")。
+// エラーが起きれば errno をセットして NULL を返す。
 string *
 net_gets(struct net *net)
 {
@@ -372,12 +375,12 @@ net_gets(struct net *net)
 			Verbose(diag, "%s: net_read=%d", __func__, n);
 			if (n < 0) {
 				Debug(diag, "%s: net_read failed: %s", __func__, strerrno());
-				return s;
+				string_free(s);
+				return NULL;
 			}
 			if (n == 0) {
 				// EOF
-				string_free(s);
-				return NULL;
+				return s;
 			}
 			net->bufpos = 0;
 			net->buflen = n;
@@ -410,12 +413,15 @@ int
 net_read(struct net *net, void *dst, uint dstsize)
 {
 	assert(net);
+	const struct diag *diag = net->diag;
 
 	// バッファにあれば先に使い切る。
 	if (net->bufpos != net->buflen) {
 		uint copylen = MIN(net->buflen - net->bufpos, dstsize);
 		memcpy(dst, net->buf + net->bufpos, copylen);
 		net->bufpos += copylen;
+		Verbose(diag, "%s: copied=%u, pos=%u/len=%u", __func__,
+			copylen, net->bufpos, net->buflen);
 		return copylen;
 	}
 
@@ -433,10 +439,10 @@ net_write(struct net *net, const void *src, uint srcsize)
 
 // 送信方向を shutdown する。
 void
-net_shutdown(struct net *net)
+net_shutdown_half(struct net *net)
 {
 	assert(net);
-	net->f_shutdown(net);
+	net->f_shutdown_half(net);
 }
 
 void
@@ -463,17 +469,16 @@ static int
 sock_connect(struct net *net, const char *host, const char *serv,
 	const struct net_opt *opt)
 {
-	struct timeval start, end, res;
+	struct timespec start, end;
 
-	gettimeofday(&start, NULL);
+	clock_gettime(CLOCK_MONOTONIC, &start);
 	net->sock = socket_connect(host, serv, opt);
 	if (net->sock < 0) {
 		return -1;
 	}
-	gettimeofday(&end, NULL);
-	timersub(&end, &start, &res);
-	Debug(net->diag, "Connected (%u msec)",
-		(uint)(res.tv_sec * 1000 + res.tv_usec / 1000));
+	clock_gettime(CLOCK_MONOTONIC, &end);
+	uint32 msec = timespec_to_msec(&end) - timespec_to_msec(&start);
+	Debug(net->diag, "Connected (%u msec)", msec);
 	return 0;
 }
 
@@ -492,7 +497,7 @@ sock_write(struct net *net, const void *src, int srcsize)
 }
 
 static void
-sock_shutdown(struct net *net)
+sock_shutdown_half(struct net *net)
 {
 	shutdown(net->sock, SHUT_WR);
 }
@@ -524,7 +529,7 @@ static int
 tls_connect(struct net *net, const char *host, const char *serv,
 	const struct net_opt *opt)
 {
-	struct timeval start, end, res;
+	struct timespec start, end;
 	const struct diag *diag = net->diag;
 	int r;
 
@@ -541,6 +546,9 @@ tls_connect(struct net *net, const char *host, const char *serv,
 		return -1;
 	}
 
+	// SSL_read/write() の WANT_READ/WRITE を向こうで処理してもらう。
+	SSL_CTX_set_mode(net->ctx, SSL_MODE_AUTO_RETRY);
+
 	if (opt->use_rsa_only) {
 		// RSA 指定なら TLSv1.2 以下を強制する。
 		// (TLSv1.3 ではそもそも RSA128-SHA とかの指定自体が存在しない)
@@ -559,7 +567,7 @@ tls_connect(struct net *net, const char *host, const char *serv,
 		return -1;
 	}
 
-	gettimeofday(&start, NULL);
+	clock_gettime(CLOCK_MONOTONIC, &start);
 
 	net->sock = socket_connect(host, serv, opt);
 	if (net->sock == -1) {
@@ -586,8 +594,7 @@ tls_connect(struct net *net, const char *host, const char *serv,
 
 	// 接続できたらログ。
 	if (__predict_false(diag_get_level(diag) >= 1)) {
-		gettimeofday(&end, NULL);
-		timersub(&end, &start, &res);
+		clock_gettime(CLOCK_MONOTONIC, &end);
 
 		SSL_SESSION *sess = SSL_get_session(net->ssl);
 		int ssl_version = SSL_SESSION_get_protocol_version(sess);
@@ -608,7 +615,7 @@ tls_connect(struct net *net, const char *host, const char *serv,
 		const SSL_CIPHER *ssl_cipher = SSL_SESSION_get0_cipher(sess);
 		const char *cipher_name = SSL_CIPHER_get_name(ssl_cipher);
 
-		uint msec = (uint)(res.tv_sec * 1000 + res.tv_usec / 1000);
+		uint32 msec = timespec_to_msec(&end) - timespec_to_msec(&start);
 		diag_print(diag, "Connected %s %s (%u msec)", ver, cipher_name, msec);
 	}
 
@@ -623,15 +630,23 @@ tls_read(struct net *net, void *dst, int dstsize)
 
 	Verbose(diag, "%s (dstsize=%u)", __func__, dstsize);
 	r = SSL_read(net->ssl, dst, dstsize);
-	if (r < 0) {
-		if (SSL_get_error(net->ssl, r) != SSL_ERROR_SYSCALL) {
+	if (r <= 0) {
+		// EOF かどうかも get_error で判断するようだ。
+		int error = SSL_get_error(net->ssl, r);
+		if (error == SSL_ERROR_ZERO_RETURN) {
+			r = 0;
+			goto done;
+		}
+		Verbose(diag, "%s r=%zd, SSL_error=%d errno=%d", __func__,
+			r, error, errno);
+		if (error != SSL_ERROR_SYSCALL) {
 			// とりあえず何かにしておく。
 			errno = EIO;
 		}
-		Verbose(diag, "%s r=%zd, errno=%d", __func__, r, errno);
-	} else {
-		Verbose(diag, "%s r=%zd", __func__, r);
+		return -1;
 	}
+ done:
+	Verbose(diag, "%s r=%zd", __func__, r);
 	return r;
 }
 
@@ -643,29 +658,40 @@ tls_write(struct net *net, const void *src, int srcsize)
 
 	Verbose(diag, "%s (srcsize=%u)", __func__, srcsize);
 	r = SSL_write(net->ssl, src, srcsize);
-	if (r < 0) {
-		if (SSL_get_error(net->ssl, r) != SSL_ERROR_SYSCALL) {
+	if (r <= 0) {
+		// EOF かどうかも get_error で判断するようだ。
+		int error = SSL_get_error(net->ssl, r);
+		if (error == SSL_ERROR_ZERO_RETURN) {
+			r = 0;
+			goto done;
+		}
+		Verbose(diag, "%s r=%zd, SSL_error=%d errno=%d", __func__,
+			r, error, errno);
+		if (error != SSL_ERROR_SYSCALL) {
 			// とりあえず何かにしておく。
 			errno = EIO;
 		}
-		Verbose(diag, "%s r=%zd, errno=%d", __func__, r, errno);
-	} else {
-		Verbose(diag, "%s r=%zd", __func__, r);
+		return -1;
 	}
+ done:
+	Verbose(diag, "%s r=%zd", __func__, r);
 	return r;
 }
 
 static void
-tls_shutdown(struct net *net)
+tls_shutdown_half(struct net *net)
 {
-	SSL_shutdown(net->ssl);
+	// SSL に shutdown(SHUT_WR) 相当の動作はないので、何もしない。
 }
 
 static void
 tls_close(struct net *net)
 {
 	if (net->ssl) {
-		SSL_shutdown(net->ssl);
+		int r = SSL_shutdown(net->ssl);
+		if (r == 0) {
+			SSL_shutdown(net->ssl);
+		}
 	}
 	// 元ソケットも閉じる。
 	sock_close(net);
@@ -694,14 +720,25 @@ static int
 socket_connect(const char *hostname, const char *servname,
 	const struct net_opt *opt)
 {
+	struct timeval tv;
+	struct timeval *tvp;
+	struct timespec now;
 	struct addrinfo hints;
 	struct addrinfo *ai;
 	struct addrinfo *ailist;
+	uint64 end_usec = 0;
 	fd_set wfds;
-	bool inprogress;
 	int fd;
 	int r;
 
+	if (__predict_false(opt->timeout_msec == 0)) {
+		tvp = NULL;
+	} else {
+		tvp = &tv;
+		clock_gettime(CLOCK_MONOTONIC, &now);
+		end_usec = timespec_to_usec(&now) + (opt->timeout_msec * 1000);
+	}
+
 	memset(&hints, 0, sizeof(hints));
 	switch (opt->address_family) {
 	 case 4:
@@ -721,9 +758,11 @@ socket_connect(const char *hostname, const char *servname,
 		return -1;
 	}
 
-	inprogress = false;
 	fd = -1;
 	for (ai = ailist; ai != NULL; ai = ai->ai_next) {
+		int val;
+		socklen_t vallen = sizeof(val);
+
 		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
 		if (fd < 0) {
 			continue;
@@ -734,55 +773,58 @@ socket_connect(const char *hostname, const char *servname,
 			goto abort_continue;
 		}
 
+		// ノンブロッキングなので connect() は EINPROGRESS を返す。
 		if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
 			break;
 		}
-		// ノンブロッキングなので connect() は EINPROGRESS を返す。
-		if (errno == EINPROGRESS) {
-			inprogress = true;
-			break;
+		if (errno != EINPROGRESS) {
+			goto abort_continue;
 		}
 
+		if (tvp != NULL) {
+			clock_gettime(CLOCK_MONOTONIC, &now);
+			uint64 now_usec = timespec_to_usec(&now);
+			if (now_usec >= end_usec) {
+				errno = ETIMEDOUT;
+				return -1;
+			}
+			uint64 timeout_usec = end_usec - now_usec;
+			tv.tv_sec  = timeout_usec / 1000000;
+			tv.tv_usec = timeout_usec % 1000000;
+		}
+		FD_ZERO(&wfds);
+		FD_SET(fd, &wfds);
+		r = select(fd + 1, NULL, &wfds, NULL, tvp);
+		if (r <= 0) {
+			goto abort_continue;
+		}
+
+		val = -1;
+		getsockopt(fd, SOL_SOCKET, SO_ERROR, &val, &vallen);
+		if (val != 0) {
+			goto abort_continue;
+		}
+
+		// ここまで来れば接続成功。
+		break;
+
  abort_continue:
 		close(fd);
 		fd = -1;
 	}
 	freeaddrinfo(ailist);
 
-	// 接続出来なかった。
+	// 一応。
 	if (fd < 0) {
 		return -1;
 	}
 
-	// ここでブロッキングに戻す。
+	// ブロッキングに戻す。
 	if (socket_setblock(fd, true) < 0) {
 		close(fd);
 		return -1;
 	}
 
-	// 接続待ちなら…
-	if (inprogress) {
-		struct timeval tv;
-		struct timeval *tvp;
-		FD_ZERO(&wfds);
-		FD_SET(fd, &wfds);
-		if (__predict_false(opt->timeout_msec == 0)) {
-			tvp = NULL;
-		} else {
-			memset(&tv, 0, sizeof(tv));
-			tv.tv_sec = opt->timeout_msec / 1000;
-			tv.tv_usec = (opt->timeout_msec % 1000) * 1000;
-			tvp = &tv;
-		}
-		r = select(fd + 1, NULL, &wfds, NULL, tvp);
-		if (r <= 0) {
-			close(fd);
-			if (r == 0) {
-				errno = ETIMEDOUT;
-			}
-			return -1;
-		}
-	}
 	return fd;
 }
 
