@@ -55,7 +55,7 @@
 #define BG_ISLIGHT()	(opt_bgtheme != BG_DARK) // 姑息な最適化
 
 // ヘッダの依存関係を減らすため。
-extern image_opt imageopt;
+extern struct image_opt imageopt;
 
 static void make_esc(char *, const char *);
 static inline void make_indent(char *, int);
@@ -80,6 +80,7 @@ void
 init_color(void)
 {
 	char url[S2EBUFSIZE];
+	char ng[S2EBUFSIZE];
 	const char *c_blue = NULL;
 	const char *c_username = NULL;
 	const char *c_renote = NULL;
@@ -87,6 +88,7 @@ init_color(void)
 	const char *c_gray = NULL;
 
 	url[0] = '\0';
+	ng[0] = '\0';
 
 	if (colormode == 1) {
 		// -c 1 なら一切エスケープシーケンスを使わない。
@@ -132,6 +134,9 @@ init_color(void)
 
 		// mlterm では 90 がグレー、97 は白。
 		c_gray = "90";
+
+		// NG。
+		snprintf(ng, sizeof(ng), "%s;%s", STRIKE, c_gray);
 	}
 
 	make_esc(style2esc[STYLE_USERNAME],	c_username);
@@ -141,6 +146,7 @@ init_color(void)
 	make_esc(style2esc[STYLE_REACTION],	c_react);
 	make_esc(style2esc[STYLE_URL],		url);
 	make_esc(style2esc[STYLE_TAG],		c_blue);
+	make_esc(style2esc[STYLE_NG],		ng);
 }
 
 static void
@@ -637,7 +643,7 @@ fetch_image(FILE *ofp, const char *img_url, uint width, uint height, bool shade)
 	FILE *ifp = NULL;
 	struct image *srcimg = NULL;
 	struct image *dstimg = NULL;
-	image_opt localopt;
+	struct image_opt localopt;
 	bool rv = false;
 
 	// dst_{width,height} は
