@@ -30,6 +30,7 @@
 
 #include "sayaka.h"
 #include "image.h"
+#include "ngword.h"
 #include <err.h>
 #include <errno.h>
 #include <getopt.h>
@@ -40,8 +41,8 @@
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 
-#define SAYAKA_VERSION "3.8.2"
-#define SAYAKA_RELDATE "2025/05/17"
+#define SAYAKA_VERSION "3.8.4"
+#define SAYAKA_RELDATE "2025/08/10"
 
 enum {
 	DEFAULT_FONT_WIDTH	= 7,
@@ -55,7 +56,7 @@ enum {
 };
 
 // ヘッダの依存関係を減らすため。
-extern image_opt imageopt;
+extern struct image_opt imageopt;
 
 static void version(void);
 static void usage(void);
@@ -64,6 +65,7 @@ static bool init(void);
 static void mkdir_if(const char *);
 static void progress(const char *);
 static void init_screen(void);
+static void init_ngword(void);
 static void invalidate_cache(void);
 static const char *get_token(const char *);
 static void signal_handler(int);
@@ -84,12 +86,13 @@ struct diag *diag_term;
 uint fontwidth;						// 使用するフォント幅   (ドット数)
 uint fontheight;					// 使用するフォント高さ (ドット数)
 uint iconsize;						// アイコンの大きさ (正方形、ドット数)
-image_opt imageopt;					// 画像関係のオプション
+struct image_opt imageopt;			// 画像関係のオプション
 uint imagesize;						// 画像の大きさ
 uint indent_cols;					// インデント1階層の桁数
 bool in_sixel;						// SIXEL 出力中か。
 struct net_opt netopt_image;		// 画像ダウンロード用ネットワークオプション
 struct net_opt netopt_main;			// メインストリーム用ネットワークオプション
+struct ngwords *ngwords;			// NG ワード集
 int opt_bgtheme;					// -1:自動判別 0:Dark 1:Light
 const char *opt_codeset;			// 出力文字コード (NULL なら UTF-8)
 static uint opt_fontwidth;			// --font 指定の幅   (指定なしなら 0)
@@ -172,6 +175,7 @@ static const struct option longopts[] = {
 	{ "sixel-or",		no_argument,		NULL,	OPT_sixel_or },
 	{ "timeout-image",	required_argument,	NULL,	OPT_timeout_image },
 	{ "token",			required_argument,	NULL,	't' },
+	{ "version",		no_argument,		NULL,	'v' },
 	{ NULL },
 };
 
@@ -232,45 +236,30 @@ main(int ac, char *av[])
 	while ((c = getopt_long(ac, av, "c:hlp:r:s:t:v", longopts, NULL)) != -1) {
 		switch (c) {
 		 case 'c':
-		 {
 			// ここは元々色数を指定しているのではなく、色モード指定。
 			// -c 2 は、画像はモノクロで、テキストはボールドのみ飾り付けを行う。
 			// -c 1 は、画像はモノクロで、テキストはボールドも含めて一切の
 			// 飾り付けを行わない (ボールドがつらい端末を救済するため)。
-			int n;
 			if (strcmp(optarg, "1") == 0) {
 				colormode = 1;
-				imageopt.color = MAKE_COLOR_FMT_GRAY(2);
-				imageopt.cdm   = 96;
-			} else if (strcmp(optarg, "2") == 0) {
-				colormode = 2;
-				imageopt.color = MAKE_COLOR_FMT_GRAY(2);
-				imageopt.cdm   = 96;
-			} else if (strcmp(optarg, "8") == 0) {
-				colormode = 8;
-				imageopt.color = COLOR_FMT_8_RGB;
-				imageopt.cdm   = 96;
-			} else if (strcmp(optarg, "16") == 0) {
-				colormode = 16;
-				imageopt.color = COLOR_FMT_16_VGA;
-			} else if (strcmp(optarg, "256") == 0) {
-				colormode = 256;
-				imageopt.color = COLOR_FMT_256_RGB332;
-			} else if (strcmp(optarg, "gray") == 0 ||
-			           strcmp(optarg, "grey") == 0) {
-				colormode = 2;
-				imageopt.color = MAKE_COLOR_FMT_GRAY(256);
-			} else if ((strncmp(optarg, "gray", 4) == 0 ||
-			            strncmp(optarg, "grey", 4) == 0   ) &&
-			           (n = stou32def(optarg + 4, -1, NULL)) != (uint32)-1 &&
-			           (2 <= n && n <= 256)) {
-				colormode = 2;
-				imageopt.color = MAKE_COLOR_FMT_GRAY(n);
+				imageopt.color = MAKE_COLOR_MODE_GRAY(2);
 			} else {
-				errx(1, "%s: invalid color mode", optarg);
+				imageopt.color = image_parse_color(optarg);
+				if (imageopt.color == COLOR_MODE_NONE) {
+					errx(1, "%s: invalid color mode", optarg);
+				}
+				switch (GET_COLOR_MODE(imageopt.color)) {
+				 case COLOR_MODE_GRAY:			colormode = 2;		break;
+				 case COLOR_MODE_8_RGB:			colormode = 8;		break;
+				 case COLOR_MODE_16_VGA:		colormode = 16;		break;
+				 case COLOR_MODE_256_RGB332:	colormode = 256;	break;
+				}
+			}
+			// 8色以下くらいだと --cdm 0.375 くらいが絶妙。
+			if (colormode <= 8) {
+				imageopt.cdm = 96;
 			}
 			break;
-		 }
 
 		 case OPT_ciphers:
 			// 今のところ "RSA" (大文字) しか指定できない。
@@ -494,7 +483,7 @@ main(int ac, char *av[])
 	}
 
 	if (cmd == CMD_STREAM || cmd == CMD_PLAY) {
-		// 表示系の初期化。
+		init_ngword();
 		init_screen();
 
 		if (cmd == CMD_STREAM) {
@@ -620,9 +609,9 @@ init(void)
 	}
 
 	// パスを用意。
-	snprintf(buf, sizeof(buf), "%s/.sayaka/", home);
+	snprintf(buf, sizeof(buf), "%s/.sayaka", home);
 	basedir = strdup(buf);
-	strlcat(buf, "cache", sizeof(buf));
+	strlcat(buf, "/cache", sizeof(buf));
 	cachedir = strdup(buf);
 	if (basedir == NULL || cachedir == NULL) {
 		return false;
@@ -682,6 +671,16 @@ progress(const char *msg)
 	}
 }
 
+// NG ワードの初期化。
+static void
+init_ngword(void)
+{
+	char filename[PATH_MAX];
+
+	snprintf(filename, sizeof(filename), "%s/ngword.json", basedir);
+	ngwords = ngword_read_file(filename, diag_format/*?*/);
+}
+
 // 表示周りの初期化。
 static void
 init_screen(void)
@@ -694,20 +693,25 @@ init_screen(void)
 	// モノクロモードなら不要。
 	if (opt_bgtheme == BG_AUTO && is_tty && colormode > 2) {
 		progress("Checking background color...");
-		opt_bgtheme = terminal_get_bgtheme();
-		switch (opt_bgtheme) {
-		 case BG_LIGHT:
-			progress("light\n");
-			break;
-		 case BG_DARK:
-			progress("dark\n");
-			break;
-		 default:
+		uint32 c = terminal_get_bgcolor();
+		if ((int32)c < 0) {
 			progress("done\n");
 			warnx("Terminal doesn't support contol sequence; "
 				"assume --light");
 			opt_bgtheme = BG_LIGHT;
-			break;
+		} else {
+			float r = (float)((c >> 16) & 0xff) / 255;
+			float g = (float)((c >>  8) & 0xff) / 255;
+			float b = (float)( c        & 0xff) / 255;
+			// グレースケールで、黒に近ければ 0、白に近ければ 1 にする。
+			// ここは背景色の明暗だけ分かればいいので細かいことは気にしない。
+			float I = (0.2126 * r) + (0.7152 * g) + (0.0722 * b);
+			opt_bgtheme = (int)(I + 0.5);
+			if (opt_bgtheme == BG_LIGHT) {
+				progress("light\n");
+			} else {
+				progress("dark\n");
+			}
 		}
 	}
 
@@ -739,8 +743,8 @@ init_screen(void)
 	init_color();
 
 	// キャッシュファイル用の色モード名。
-	switch (imageopt.color & COLOR_FMT_MASK) {
-	 case COLOR_FMT_GRAY:
+	switch (GET_COLOR_MODE(imageopt.color)) {
+	 case COLOR_MODE_GRAY:
 	 {
 		uint grayscale = GET_COLOR_COUNT(imageopt.color);
 		if (grayscale == 2) {
@@ -750,13 +754,13 @@ init_screen(void)
 		}
 		break;
 	 }
-	 case COLOR_FMT_8_RGB:
+	 case COLOR_MODE_8_RGB:
 		strlcpy(colorname, "8", sizeof(colorname));
 		break;
-	 case COLOR_FMT_16_VGA:
+	 case COLOR_MODE_16_VGA:
 		strlcpy(colorname, "16", sizeof(colorname));
 		break;
-	 case COLOR_FMT_256_RGB332:
+	 case COLOR_MODE_ADAPTIVE:
 		strlcpy(colorname, "256", sizeof(colorname));
 		break;
 	 default:
@@ -775,9 +779,9 @@ invalidate_cache(void)
 {
 	char cmd[1024];
 
-	// アイコンは1か月分くらいか。
+	// アイコンは1週間分くらいか。
 	snprintf(cmd, sizeof(cmd),
-		"find %s -name icon-\\* -type f -atime +30 -exec rm {} +", cachedir);
+		"find %s -name icon-\\* -type f -atime +7 -exec rm {} +", cachedir);
 	if (system(cmd) < 0) {
 		warn("system(find icon)");
 	}
