@@ -36,15 +36,17 @@
 #include <getopt.h>
 #include <signal.h>
 #include <string.h>
+#include <time.h>
 #include <unistd.h>
-#include <sys/time.h>
+#include <sys/ioctl.h>
 
-#define SIXELV_VERSION "3.8.2"
-#define SIXELV_RELDATE "2025/05/17"
+#define SIXELV_VERSION "3.8.4"
+#define SIXELV_RELDATE "2025/08/10"
 
 typedef enum {
 	OUTPUT_FORMAT_SIXEL,
 	OUTPUT_FORMAT_BMP,
+	OUTPUT_FORMAT_ASCII,
 } OutputFormat;
 
 static void version(void);
@@ -58,6 +60,8 @@ static void signal_handler(int);
 static struct diag *diag_image;
 static struct diag *diag_net;
 static struct diag *diag_sixel;
+static uint fontwidth;				// フォント幅 (ドット数)
+static uint fontheight;				// フォント高さ (ドット数)
 static bool show_filename;			// 画像の前にファイル名を表示
 static bool ignore_error;			// true ならエラーでも次ファイルを処理
 static FILE *ofp;					// 出力中のストリーム
@@ -65,15 +69,16 @@ static bool opt_blurhash_nearest;	// Blurhash を最近傍補間する
 static ResizeAxis opt_resize_axis;
 static uint opt_width;
 static uint opt_height;
+static int  opt_page;
 static bool opt_profile;			// プロファイル
 static const char *output_filename;	// 出力ファイル名。NULL なら stdout
 static OutputFormat output_format;	// 出力形式
-static image_opt imageopt;
+static struct image_opt imageopt;
 static struct net_opt netopt;
 
-#define PROF(tv)	do {		\
-	if (opt_profile)			\
-		gettimeofday(tv, NULL);	\
+#define PROF(tv)	do {					\
+	if (opt_profile)						\
+		clock_gettime(CLOCK_MONOTONIC, tv);	\
 } while (0)
 
 const char progname[] = "sixelv";
@@ -123,6 +128,7 @@ static const struct option longopts[] = {
 	{ "ipv6",			no_argument,		NULL,	OPT_ipv6 },
 	{ "list",			no_argument,		NULL,	OPT_list },
 	{ "output-format",	required_argument,	NULL,	'O' },
+	{ "page",			required_argument,	NULL,	'p' },
 	{ "profile",		no_argument,		NULL,	OPT_profile },
 	{ "reduction",		required_argument,	NULL,	'r' },
 	{ "resize-axis",	required_argument,	NULL,	OPT_resize_axis },
@@ -135,6 +141,7 @@ static const struct option longopts[] = {
 };
 
 static const struct optmap map_output_format[] = {
+	{ "ascii",		OUTPUT_FORMAT_ASCII },
 	{ "bmp",		OUTPUT_FORMAT_BMP },
 	{ "sixel",		OUTPUT_FORMAT_SIXEL },
 	{ NULL },
@@ -207,7 +214,7 @@ main(int ac, char *av[])
 	output_filename = NULL;
 	output_format = OUTPUT_FORMAT_SIXEL;
 
-	while ((c = getopt_long(ac, av, "c:d:h:iO:o:r:vw:",
+	while ((c = getopt_long(ac, av, "c:d:h:iO:o:p:r:vw:",
 					longopts, NULL)) != -1)
 	{
 		switch (c) {
@@ -216,31 +223,11 @@ main(int ac, char *av[])
 			break;
 
 		 case 'c':
-		 {
-			int n;
-			if (strcmp(optarg, "2") == 0) {
-				imageopt.color = MAKE_COLOR_FMT_GRAY(2);
-			} else if (strcmp(optarg, "8") == 0) {
-				imageopt.color = COLOR_FMT_8_RGB;
-			} else if (strcmp(optarg, "16") == 0) {
-				imageopt.color = COLOR_FMT_16_VGA;
-			} else if (strcmp(optarg, "256") == 0) {
-				imageopt.color = COLOR_FMT_256_RGB332;
-			} else if (strcmp(optarg, "xterm256") == 0) {
-				imageopt.color = COLOR_FMT_256_XTERM;
-			} else if (strcmp(optarg, "gray") == 0 ||
-			           strcmp(optarg, "grey") == 0) {
-				imageopt.color = MAKE_COLOR_FMT_GRAY(256);
-			} else if ((strncmp(optarg, "gray", 4) == 0 ||
-			            strncmp(optarg, "grey", 4) == 0   ) &&
-			           (n = stou32def(optarg + 4, -1, NULL)) != (uint32)-1 &&
-			           (2 <= n && n <= 256)) {
-				imageopt.color = MAKE_COLOR_FMT_GRAY(n);
-			} else {
+			imageopt.color = image_parse_color(optarg);
+			if (imageopt.color == COLOR_MODE_NONE) {
 				errx(1, "%s: invalid color mode", optarg);
 			}
 			break;
-		 }
 
 		 case OPT_cdm:
 		 {
@@ -286,7 +273,10 @@ main(int ac, char *av[])
 			if (f < 0 || f > 2) {
 				errx(1, "invalid gain");
 			}
-			imageopt.gain = (uint)(f * 256);
+			imageopt.gain = (int)(f * 256);
+			if (imageopt.gain == 256) {
+				imageopt.gain = -1;
+			}
 			break;
 		 }
 
@@ -338,6 +328,13 @@ main(int ac, char *av[])
 			}
 			break;
 
+		 case 'p':
+			opt_page = stou32def(optarg, -1, NULL);
+			if (opt_page == -1) {
+				errx(1, "invalid page: %s", optarg);
+			}
+			break;
+
 		 case OPT_profile:
 			opt_profile = true;
 			break;
@@ -403,6 +400,32 @@ main(int ac, char *av[])
 			"-o <output_filename> cannot be used with multiple input file.");
 	}
 
+	// 出力形式が ASCII でフォントサイズ未指定ならここでフォントサイズを取得。
+	// 出力先が端末でなければ適当な値を代入する。
+	if (output_format == OUTPUT_FORMAT_ASCII &&
+			(fontwidth == 0 || fontheight == 0))
+	{
+		struct winsize ws;
+		int r = ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws);
+		if (r == 0) {
+			if (ws.ws_col != 0) {
+				fontwidth = ws.ws_xpixel / ws.ws_col;
+			}
+			if (ws.ws_row != 0) {
+				fontheight = ws.ws_ypixel / ws.ws_row;
+			}
+		}
+
+		// 取得できなければ適当な値を入れておく。
+		if (fontwidth == 0) {
+			fontwidth = 7;
+		}
+		if (fontheight == 0) {
+			fontheight = 14;
+		}
+	}
+
+	signal(SIGPIPE, SIG_IGN);
 	if (output_format == OUTPUT_FORMAT_SIXEL) {
 		signal(SIGINT, signal_handler);
 	}
@@ -449,12 +472,14 @@ usage(void)
 	fprintf(stderr,
 		"usage: %s [<options...>] [-|<file|url...>]\n", progname);
 	fprintf(stderr,
-"  -c <color>      : Color mode. 2, 8, 16, 256 or gray[2..256] (default:256)\n"
+"  -c <color>      : Color mode. 2, 8, 16, 256, gray[2..256] (default:256)\n"
+"                    (See --help-all for more details)\n"
 "  -w <width>      : Resize width to <width> pixel\n"
 "  -h <height>     : Resize height to <height> pixel\n"
 "  -r <method>     : Reduction method, none(simple) or high (default:high)\n"
-"  -O <fmt>        : Output format, bmp or sixel (default: sixel)\n"
+"  -O <fmt>        : Output format, ascii, bmp or sixel (default: sixel)\n"
 "  -o <filename>   : Output filename, '-' means stdout (default: -)\n"
+"  -p <page>       : Specify the page(frame). (animated GIF/WebP only)\n"
 "  -v              : Show input filename\n");
 
 	static const char * const opts[] = {
@@ -485,13 +510,15 @@ help_all(void)
 		"usage: %s [<options...>] [-|<file|url...>]\n", progname);
 	fprintf(stderr,
 "  -c,--color=<colormode> : Set color mode (default:256)\n"
-"     256      : Fixed 256 colors (MSX SCREEN8 compatible palette)\n"
-"     xterm256 : Fixed 256 colors (xterm compatible palette)\n"
-"     16       : Fixed ANSI compatible 16 colors\n"
-"     8        : Fixed RGB 8 colors\n"
-"     2        : Monochrome (2-level grayscale)\n"
-"     gray[<n>]: (2..256) shades of grayscale. If <n> is omitted, 256 is used\n"
-"                'gray2' is a synonym for '2'\n"
+"     256          : A synonym for 'adaptive256'\n"
+"     adaptive[<n>]: Adaptive palette (n=8..256). If <n> is omitted, 256 is used\n"
+"     fixed256     : Fixed 256 colors (MSX SCREEN8 compatible palette)\n"
+"     xterm256     : Fixed 256 colors (xterm compatible palette)\n"
+"     16           : Fixed ANSI compatible 16 colors\n"
+"     8            : Fixed RGB 8 colors\n"
+"     2            : Monochrome (2-level grayscale)\n"
+"     gray[<n>]    : (2..256) shades of grayscale. 'gray2' is a synonym for '2'.\n"
+"                    If <n> is omitted, 256 is used\n"
 "  -w,--width=<width>     : Resize width to <width> pixel\n"
 "  -h,--height=<height>   : Resize height to <height> pixel\n"
 "  --resize-axis=<axis>   : Set an origin axis for resizing (default:both)\n"
@@ -516,8 +543,9 @@ help_all(void)
 "  --gain=<gain>          : Set output gain between 0.0 and 2.0 (default:1.0)\n"
 "  --help-all             : This help\n"
 "  --list                 : Show supported filetype and decoder list\n"
-"  -O,--output-format=<fmt> : bmp or sixel (default:sixel)\n"
+"  -O,--output-format=<fmt> : ascii, bmp or sixel (default:sixel)\n"
 "  -o <filename>          : Output filename, '-' means stdout (default:-)\n"
+"  -p,--page=<page>       : Specify the page(frame). (animated GIF/WebP only)\n"
 "  --sixel-or             : Output SIXEL by OR-mode\n"
 "  --sixel-transbg        : Make SIXEL background transparent\n"
 "  --suppress-palette     : Suppress output of SIXEL palette definition\n"
@@ -549,13 +577,13 @@ do_file(const char *infile)
 	image_read_hint hint;
 	uint dst_width;
 	uint dst_height;
-	struct timeval load_start;
-	struct timeval load_end;
-	struct timeval cvt_start;
-	struct timeval reduct_start;
-	struct timeval reduct_end;
-	struct timeval sixel_start;
-	struct timeval sixel_end;
+	struct timespec load_start;
+	struct timespec load_end;
+	struct timespec cvt_start;
+	struct timespec reduct_start;
+	struct timespec reduct_end;
+	struct timespec sixel_start;
+	struct timespec sixel_end;
 
 	infilename = infile;
 	if (infile == NULL) {
@@ -619,12 +647,14 @@ do_file(const char *infile)
 
 	if (loader_idx >= 0) {
 		// 読み込み。
-		// (この hint は libjpeg の scaling hint のことで、これをもとに
-		// 1/8 とかで読み込んだものが srcimg->{width,height} になる)
+		// この hint は libjpeg の scaling hint のことで、これをもとに
+		// 1/8 とかで読み込んだものが srcimg->{width,height} になる。
+		// ASCII の時には hint サイズも加工したほうが効率はいいがとりあえず。
 		memset(&hint, 0, sizeof(hint));
 		hint.axis   = opt_resize_axis;
 		hint.width  = opt_width;
 		hint.height = opt_height;
+		hint.page   = opt_page;
 		srcimg = image_read(pstream, loader_idx, &hint, diag_image);
 		if (srcimg) {
 			// 得られた画像サイズと引数指定から、いい感じにサイズを決定。
@@ -644,10 +674,22 @@ do_file(const char *infile)
 		goto abort;
 	}
 
+	if (output_format == OUTPUT_FORMAT_ASCII) {
+		// ここでピクセルサイズを桁数行数に変更。
+		dst_width  = howmany(dst_width, fontwidth);
+		dst_height = howmany(dst_height, fontheight);
+	}
+
 	Debug(diag_image,
 		"InputSize=(%u, %u) OutputSize=(%u, %u) OutputColor=%s",
 		srcimg->width, srcimg->height, dst_width, dst_height,
-		colorformat_tostr(imageopt.color));
+		colormode_tostr(imageopt.color));
+
+	if (dst_width == 0 || dst_height == 0) {
+		warnx("%s: Output size (%u, %u) is too small",
+			infilename, dst_width, dst_height);
+		goto abort;
+	}
 
 	PROF(&cvt_start);
 	image_convert_to16(srcimg);
@@ -662,6 +704,13 @@ do_file(const char *infile)
 
 	PROF(&reduct_end);
 
+	if (GET_COLOR_MODE(imageopt.color) == COLOR_MODE_ADAPTIVE) {
+		Debug(diag_image, "AdaptivePalette InputColors=%u%s OutputColors=%u",
+			srcimg->palette_count,
+			(srcimg->palette_count > 256 ? "/32768" : ""),
+			resimg->palette_count);
+	}
+
 	// 出力先をオープン。
 	if (output_filename == NULL) {
 		ofp = stdout;
@@ -676,35 +725,45 @@ do_file(const char *infile)
 	PROF(&sixel_start);
 
 	// 書き出し。
-	if (output_format == OUTPUT_FORMAT_SIXEL) {
+	switch (output_format) {
+	 case OUTPUT_FORMAT_SIXEL:
 		image_sixel_write(ofp, resimg, &imageopt, diag_sixel);
-	} else {
+		break;
+	 case OUTPUT_FORMAT_BMP:
 		if (image_bmp_write(ofp, resimg, diag_image) == false) {
 			goto abort;
 		}
+		break;
+	 case OUTPUT_FORMAT_ASCII:
+		if (image_ascii_write(ofp, resimg, &imageopt, diag_image) == false) {
+			goto abort;
+		}
+		break;
 	}
 	fflush(ofp);
 
 	PROF(&sixel_end);
 
 	if (opt_profile) {
-		struct timeval load_res;
-		struct timeval cvt_res;
-		struct timeval reduct_res;
-		struct timeval sixel_res;
-		timersub(&load_end, &load_start, &load_res);
-		timersub(&reduct_start, &cvt_start, &cvt_res);
-		timersub(&reduct_end, &reduct_start, &reduct_res);
-		timersub(&sixel_end, &sixel_start, &sixel_res);
-		float ltime = load_res.tv_sec   + (float)load_res.tv_usec / 1000000;
-		float ctime = cvt_res.tv_sec    + (float)cvt_res.tv_usec / 1000000;
-		float rtime = reduct_res.tv_sec + (float)reduct_res.tv_usec / 1000000;
-		float stime = sixel_res.tv_sec  + (float)sixel_res.tv_usec / 1000000;
+		uint64 load_usec;
+		uint64 cvt_usec;
+		uint64 reduct_usec;
+		uint64 sixel_usec;
+		load_usec = timespec_to_usec(&load_end) - timespec_to_usec(&load_start);
+		cvt_usec = timespec_to_usec(&reduct_end) - timespec_to_usec(&cvt_start);
+		reduct_usec = timespec_to_usec(&reduct_end)
+											- timespec_to_usec(&reduct_start);
+		sixel_usec = timespec_to_usec(&sixel_end)
+											- timespec_to_usec(&sixel_start);
+		float ltime = (float)load_usec   / 1000;
+		float ctime = (float)cvt_usec    / 1000;
+		float rtime = (float)reduct_usec / 1000;
+		float stime = (float)sixel_usec  / 1000;
 		diag_print(diag_image,
 			"Load(+IO) %4.1f, Cvt %4.1f, Reduct %4.1f, %s(+IO) %4.1f msec",
-			ltime * 1000, ctime * 1000, rtime * 1000,
+			ltime, ctime, rtime,
 			(output_format == OUTPUT_FORMAT_SIXEL ? "SIXEL" : "Write"),
-			stime * 1000);
+			stime);
 	}
 
 	rv = true;
