@@ -31,6 +31,7 @@
 #include "sayaka.h"
 #include <errno.h>
 #include <string.h>
+#include <time.h>
 #include <sys/select.h>
 #include <sys/time.h>
 
@@ -210,6 +211,11 @@ wsclient_connect(struct wsclient *ws, const char *url,
 	// 応答の1行目を受信。
 	response = net_gets(ws->net);
 	if (response == NULL) {
+		Debug(diag, "%s: %s", __func__, strerrno());
+		rv  = -1;
+		goto abort;
+	}
+	if (string_len(response) == 0) {
 		Debug(diag, "%s: Unexpected EOF while reading response header",
 			__func__);
 		rv = 0;
@@ -219,8 +225,22 @@ wsclient_connect(struct wsclient *ws, const char *url,
 	Trace(diag, "--> |%s|", string_get(response));
 
 	// 残りの行は今のところ使ってないので読み捨てる。
-	string *recvhdr;
-	while ((recvhdr = net_gets(ws->net)) != NULL) {
+	for (;;) {
+		string *recvhdr;
+
+		recvhdr = net_gets(ws->net);
+		if (recvhdr == NULL) {
+			Debug(diag, "%s: %s", __func__, strerrno());
+			rv = -1;
+			goto abort;
+		}
+		if (string_len(recvhdr) == 0) {
+			Debug(diag, "%s: Unexpected EOF while reading header", __func__);
+			string_free(recvhdr);
+			rv = 0;
+			goto abort;
+		}
+
 		string_rtrim_inplace(recvhdr);
 		bool newline = (string_len(recvhdr) == 0);
 		Trace(diag, "--> |%s|", string_get(recvhdr));
@@ -273,9 +293,10 @@ int
 wsclient_process(struct wsclient *ws)
 {
 	const struct diag *diag = ws->diag;
-	struct timeval timeout;
-	struct timeval now;
-	struct timeval end;
+	struct timespec now;
+	uint64 now_usec;
+	uint64 end_usec;
+	uint64 timeout_usec;
 	int fd;
 	int rv = 1;
 	int r;
@@ -294,26 +315,28 @@ wsclient_process(struct wsclient *ws)
 
 	// キープアライブのため一定時間だけ受信を待つ。
 	fd = net_get_fd(ws->net);
-	timeout.tv_sec = 30;
-	timeout.tv_usec = 0;
-	gettimeofday(&now, NULL);
-	timeradd(&now, &timeout, &end);
+	timeout_usec = 30 * 1000 * 1000;
+	clock_gettime(CLOCK_MONOTONIC, &now);
+	now_usec = timespec_to_usec(&now);
+	end_usec = now_usec + timeout_usec;
 	for (;;) {
+		struct timeval tv;
 		fd_set rfds;
 		FD_ZERO(&rfds);
 		FD_SET(fd, &rfds);
 
-		gettimeofday(&now, NULL);
-		if (((end.tv_sec != now.tv_sec)
-			? (end.tv_sec - now.tv_sec) : (end.tv_usec - now.tv_usec)) <= 0)
-		{
+		clock_gettime(CLOCK_MONOTONIC, &now);
+		now_usec = timespec_to_usec(&now);
+		if (now_usec >= end_usec) {
 			// 一定時間何も起きなかったので PING を投げる。
 			wsclient_send_ping(ws);
 			return 1;
 		}
 
-		timersub(&end, &now, &timeout);
-		r = select(fd + 1, &rfds, NULL, NULL, &timeout);
+		timeout_usec = end_usec - now_usec;
+		tv.tv_sec  = timeout_usec / 1000000;
+		tv.tv_usec = timeout_usec % 1000000;
+		r = select(fd + 1, &rfds, NULL, NULL, &tv);
 		if (r < 0) {
 			if (errno == EINTR) {
 				continue;
@@ -356,7 +379,7 @@ wsclient_process(struct wsclient *ws)
 	// ペイロードを全部読み込めているか。
 	if (ws->buflen - pos < datalen) {
 		// 足りなければ次のフレームを待つ。
-		Debug(diag, "%s: wait more data: filled=%u < datalen=%u", __func__,
+		Trace(diag, "%s: wait more data: filled=%u < datalen=%u", __func__,
 			ws->buflen - pos, datalen);
 		return 1;
 	}
@@ -367,11 +390,11 @@ wsclient_process(struct wsclient *ws)
 	// opcode ごとの処理。
 	// バイナリフレームは未対応。
 	if (opcode == WS_OPCODE_PING) {
-		Debug(diag, "%s: PING len=%u recved", __func__, datalen);
+		Trace(diag, "%s: PING len=%u recved", __func__, datalen);
 		wsclient_send_pong(ws);
 		rv = 1;
 	} else if (opcode == WS_OPCODE_PONG) {
-		Debug(diag, "%s: PONG len=%u recved", __func__, datalen);
+		Trace(diag, "%s: PONG len=%u recved", __func__, datalen);
 		rv = 1;
 	} else if (opcode == WS_OPCODE_CLOSE) {
 		Debug(diag, "%s: CLOSE", __func__);
@@ -425,7 +448,7 @@ static inline void
 wsclient_send_ping(struct wsclient *ws)
 {
 	wsclient_send(ws, WS_OPCODE_PING, NULL, 0);
-	Debug(ws->diag, "%s", __func__);
+	Trace(ws->diag, "%s", __func__);
 }
 
 // PONG 応答を送信する。
@@ -433,7 +456,7 @@ static inline void
 wsclient_send_pong(struct wsclient *ws)
 {
 	wsclient_send(ws, WS_OPCODE_PONG, NULL, 0);
-	Debug(ws->diag, "%s", __func__);
+	Trace(ws->diag, "%s", __func__);
 }
 
 // WebSocket フレームを送信する。
@@ -588,7 +611,18 @@ testhttp(const struct diag *diag, int ac, char *av[])
 
 	// HTTP 応答を受信して表示。
 	string *buf;
-	while ((buf = net_gets(net)) != NULL) {
+	for (;;) {
+		buf = net_gets(net);
+		if (buf == NULL) {
+			printf("%s", strerrno());
+			break;
+		}
+		if (string_len(buf) == 0) {
+			printf("Unexpected EOF\n");
+			string_free(buf);
+			break;
+		}
+
 		string_rtrim_inplace(buf);
 		bool end = (string_len(buf) == 0);
 		printf("%s\n", string_get(buf));
